#include <ESP8266WiFi.h>
#include <WiFiUDP.h>
#include "ln_opc.h"
#include <EEPROM.h>
#include <Adafruit_PWMServoDriver.h>
#include <Adafruit_MCP23017.h>

Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver();
Adafruit_MCP23017 mcp1;

WiFiServer server(80);

WiFiUDP Udp;

boolean serterm=true;

char ssid[] = "BR218";  
char pass[] = "12345678";
byte bcflags[4];

byte a218dirspeed;
byte a218db4;
byte a218dir;
byte tempdir;

int ENA=D2;
int IN1=D5;
int IN2=D6;
int ENB=D7;
int IN3=D12;
int IN4=D13;




static IPAddress Ip(192, 168, 0, 111); //default IP
static IPAddress Gw(192, 168, 0, 1); //default Gateway
static IPAddress Sb(255, 255, 255, 0); //default Subnet

const int led = 16;

const int port=21105;
const int maxClients=16;

byte mybuffer[1024];

int lesezeiger;
int schreibzeiger;
boolean lz0;
boolean sz0;

typedef struct   
{
  IPAddress IP;
  boolean inuse;
} listofIP;
listofIP iplist[maxClients];

int ipcounter;


typedef struct
{
  unsigned char ucSTAT;   // slot status
  unsigned char ucADR;    // least significant 7 bits or short address
  unsigned char ucSPD;    // Speed
  unsigned char ucDIRF;   // direction & lower functions
  unsigned char ucSS2;    // slot status 2
  unsigned char ucADR2;   // most significant 7 bit or zero for short address
  unsigned char ucSND;    // the good old "upper" function group
  unsigned char ucID1;
  unsigned char ucID2;
}SSlotData_t;

const int SLOTMAX=120;
static SSlotData_t SlotTabelle[SLOTMAX];


byte sendSRN[5]={0x08, 0x00, 0x10, 0x00, 0x01};
byte sendVers[9]={0x09, 0x00, 0x40, 0x00, 0x63, 0x21, 0x30, 0x12, 0x60};
byte trackOff[7]={0x07, 0x00, 0x40, 0x00, 0x61, 0x00, 0x61};
byte trackOn[7]={0x07, 0x00, 0x40, 0x00, 0x61, 0x01, 0x60};
byte sendStatus[8]={0x08,0x00, 0x40,0x00, 0x62, 0x22, 0x00, (0x40^0x62^0x22)};
byte progmode[7]={0x07, 0x00, 0x40, 0x00, 0x61, 0x02, 0x63};
byte shortcircuit[7]={0x07, 0x00, 0x40, 0x00, 0x61, 0x08, 0x69};
byte unknown[7]={0x07, 0x00, 0x40, 0x00, 0x61, 0x82, 0xe3};
byte notstop[7]={0x07, 0x00, 0x40, 0x00, 0x81, 0x00, 0x81};
byte firmware[9]={0x09, 0x00, 0x40, 0x00, 0xf3, 0x0a, 0x00, 0x01, (0xf3^0x0a^0x01)};



void motor_ini(){
  pinMode(IN1, OUTPUT);
  pinMode(IN2, OUTPUT);
  pinMode(IN3, OUTPUT);
  pinMode(IN4, OUTPUT);
}

void motor_run(int ldirspeed){
  if (ldirspeed&(1<<7)){
    digitalWrite(IN1, HIGH);
    digitalWrite(IN2, LOW);
    digitalWrite(IN3, HIGH);
    digitalWrite(IN4, LOW);
  }else{
    digitalWrite(IN1, LOW);
    digitalWrite(IN2, HIGH);
    digitalWrite(IN3, LOW);
    digitalWrite(IN4, HIGH);
  }
  ldirspeed&=~(1<<7);
  analogWrite(ENA,(ldirspeed*8));
  analogWrite(ENB,(ldirspeed*8));
}


void led_an(int led){
  pwm.setPin(led, 4096);
}

void led_aus(int led){
  pwm.setPin(led, 0);
}


void setup() {
  lz0=false;
  sz0=false;
  Serial.begin(115200);
  Serial.println("Debug-Port");
  lesezeiger=0;
  schreibzeiger=0;
  bcflags[3]=1;
  ipcounter=0;
  pinMode(led, OUTPUT);
  digitalWrite(led, 1);
    Serial.println(ssid);
    Serial.println(pass);
  Serial.println("Z21-Imitat");
  WiFi.mode(WIFI_AP);  //nur Access-Point

  WiFi.softAP(ssid, pass);    //Start AcessPoint
  WiFi.softAPConfig(Ip, Gw, Sb);  //set the IP for Z21
  Udp.begin(21105);
  digitalWrite(led, 0);
  delay(2000);
  IPAddress myIP = WiFi.softAPIP();
   Serial.print("AP IP address: ");
   Serial.println(myIP);
 pwm.begin();
  pwm.setPWMFreq(1000);
  Wire.setClock(400000);
  mcp1.begin(0);
  mcp1.pinMode(15, INPUT);
  mcp1.pinMode(14, INPUT);
  mcp1.pullUp(15, HIGH);
  mcp1.pullUp(14, HIGH);
  for (int i=0; i<16; i++) led_an(i);
  delay(3000);
  for (int i=15; i>=0; i--) led_aus(i);
  delay(3000);
  
  motor_ini();
}




void AsendMessage(rwSlotDataMsg margs){
  byte zahl[4]={
  byte(margs.command),
  byte(margs.slot),
  byte(margs.adr),    // least significant 7 bits or short address
  byte(margs.adr2),   // most significant 7 bit or zero for short address
  };
   for (int i=0; i < ipcounter; i++){
    if (iplist[i].inuse){
      
      Udp.beginPacket(iplist[i].IP, port);
      Udp.write(zahl, 4);
      Udp.endPacket();
    }}  
}

void sendMessage(byte daten[]){
  int laenge=sizeof(daten);
  for (int i=0; i < ipcounter; i++){
    if (iplist[i].inuse){
     // if (serterm) 
      Serial.println("Senden...");
      Udp.beginPacket(iplist[i].IP, port);
      Udp.write(daten, daten[0]);
      Udp.endPacket();
    }
  }
}

void sendBCFlags(){
  byte sendbcflags[8]={0x08, 0x00, 0x51, 0x00, bcflags[0], bcflags[1], bcflags[2], bcflags[3]};
  sendMessage(sendbcflags);
}


void sendLocoInfo(int lok){
  sendMessage(unknown);
  
}



void update_speed(int slot, int spd){
  SlotTabelle[slot].ucSPD=spd;
  sendLocoInfoSL(slot);
}

void update_dirf(int slot, int dirf){
  SlotTabelle[slot].ucDIRF=dirf;
  sendLocoInfoSL(slot);
}

void update_slot(rwSlotDataMsg anf){
  SlotTabelle[anf.slot].ucSTAT= anf.stat;   // slot status
  SlotTabelle[anf.slot].ucADR=anf.adr;    // least significant 7 bits or short address
  SlotTabelle[anf.slot].ucSPD=anf.spd;  // Speed
  SlotTabelle[anf.slot].ucDIRF=anf.dirf;   // direction & lower functions
  SlotTabelle[anf.slot].ucSS2=anf.ss2;    // slot status 2
  SlotTabelle[anf.slot].ucADR2=anf.adr2;   // most significant 7 bit or zero for short address
  SlotTabelle[anf.slot].ucSND=anf.snd;    // the good old "upper" function group
  SlotTabelle[anf.slot].ucID1=anf.id1;
  SlotTabelle[anf.slot].ucID2=anf.id2;
}


void sendLocoInfoSL(int myslot){
  int lokadresse=(SlotTabelle[myslot].ucADR2*128)+SlotTabelle[myslot].ucADR;
  byte msb=0;;
  byte lsb=0;
  
  if (lokadresse>128){
             msb=  (0xc0|(lokadresse/256));
             lsb= (lokadresse%256);
        } else lsb=byte(lokadresse);
  byte db2=(SlotTabelle[myslot].ucSTAT&=~0xf8);
  byte temp1=SlotTabelle[myslot].ucDIRF;
  byte db3=SlotTabelle[myslot].ucSPD+((temp1=~0xdf)<<2);
  byte db4=SlotTabelle[myslot].ucDIRF; //Serial1.print("myslot.ucDIRF"); Serial1.print(db4);
  byte db5=(SlotTabelle[myslot].ucSND&=~0xf0);;
  byte db6=0;
  byte db7=0;
  byte mxor=0xef^msb^lsb^db2^db3^db4^db5^db6^db7;
  byte sLiSL[14]={0x11,0x00,0x40,0x00,0xef,msb,lsb,db2,db3,db4,db5, db6, db7,mxor};
  sendMessage(sLiSL);
}

void sendLocoInfo(byte adr_msb, byte adr_lsb, byte dirspeed, byte func1, byte func2, byte func3, byte func4){
  byte msb=adr_msb;
  byte lsb=adr_lsb;
  byte db2=4; //(SlotTabelle[myslot].ucSTAT&=~0xf8);
  byte db3=dirspeed; //SlotTabelle[myslot].ucSPD+((temp1=~0xdf)<<2);
  byte db4=func1; //SlotTabelle[myslot].ucDIRF; //Serial1.print("myslot.ucDIRF"); Serial1.print(db4);
  byte db5=func2; //(SlotTabelle[myslot].ucSND&=~0xf0);;
  byte db6=func3;
  byte db7=func4;
  byte mxor=0xef^msb^lsb^db2^db3^db4^db5^db6^db7;
  byte sLiSL[14]={0x11,0x00,0x40,0x00,0xef,msb,lsb,db2,db3,db4,db5, db6, db7,mxor};
  sendMessage(sLiSL);
}



void parseMessage(byte toParse[], byte packetSize){
    for (int i = 0; i<packetSize; i++){
     Serial.print(toParse[i],HEX);
     Serial.print(" ");
    }
  Serial.println();
´  if ((toParse[0]==0x07)&&(toParse[2]==0x40)&&(toParse[4]==0x21)){
      if (toParse[5]==0x80){
        Serial.println("GPOFF");
        sendMessage(trackOff);
      }
      if (toParse[5]==0x81){
        Serial.println("GPOn");
        sendMessage(trackOn);
      }
    }
  if ((toParse[0]==0x04)&&(toParse[1]==0)&&(toParse[2]==0x10)&&(toParse[4]==0)){
    sendMessage(sendSRN);  
  }
  if ((toParse[0]==0x04)&&(toParse[1]==0)&&(toParse[2]==0x51)&&(toParse[4]==0)){
    sendBCFlags();  
  }
  if ((toParse[0]==0x07)&&(toParse[2]==0x40)&&(toParse[4]==0x21)&&(toParse[5]==0x21)){
    sendMessage(sendVers);  
  } 
  if ((toParse[0]==0x07)&&(toParse[2]==0x40)&&(toParse[4]==0x21)&&(toParse[5]==0x24)){
    sendMessage(sendStatus);  
  }
  if ((toParse[0]==0x06)&&(toParse[2]==0x40)&&(toParse[4]==0x80)&&(toParse[5]==0x80)){
    sendMessage(notstop);
  }
  if ((toParse[0]==0x07)&&(toParse[2]==0x40)&&(toParse[4]==0xf1)&&(toParse[5]==0x0a)){
    sendMessage(firmware);  
  }
  if ((toParse[2]==0x040)&&(toParse[3]==0x00)&&(toParse[4]==0x43)){
    int infoweiche=(toParse[5]<<8)+toParse[6];
    sendTurnoutInfo(infoweiche);
  }
  if ((toParse[2]==0x040)&&(toParse[3]==0x00)&&(toParse[4]==0x53)){
    int setweiche=(toParse[5]<<8)+toParse[6];
    Serial.print("Weiche: ");
    Serial.println(setweiche);
    byte turnoutX=toParse[7];
    if (turnoutX&(1<<3)){
      if (turnoutX&(1<<0)){
        turnoutX=2;
      }else turnoutX=1;
    }
    Serial.print("Dir: ");
    Serial.println(turnoutX);
    sendTurnout(setweiche, turnoutX);
  }
  if ((toParse[2]==0x040)&&(toParse[3]==0x00)&&(toParse[4]==0xe3)&&(toParse[5]==0xf0)){
    int infoLok=toParse[7];
    if(toParse[6]>192) infoLok=((toParse[6]-192)*256)+toParse[7];
    Serial.print("Lok: ");
    Serial.println(infoLok);
    locoAdrMsg nlok;
    nlok.command=0xbf;
    nlok.adr_hi=infoLok/128;
    nlok.adr_lo=infoLok%128;
    sendLocoInfoSL(infoLok);
  }
  if ((toParse[2]==0x040)&&(toParse[3]==0x00)&&(toParse[4]==0xe4)&&(toParse[5]==0xF8)){
    int fLok=toParse[7];
    if(toParse[6]>192) fLok=((toParse[6]-192)*256)+toParse[7];
    Serial.print("Funktion roh: ");
    Serial.println(toParse[8]);
    boolean an=false;
    int func=toParse[8];
    if (func &(1<<6)) an = true;
    func &=~(1<<7);
    func &=~(1<<6);
    Serial.print("Funktion: ");
    Serial.println(func);
    int db4=0;
    if (func==0){
     if(an) db4|=(1<<4);
    if(an){ 
      led_an(1);
      led_an(2);
      led_an(3);
    }
     if(!an){
      led_aus(1);
      led_aus(2);
      led_aus(3);
     }
    }
    if (func==4){
      db4|=(1<<3);
   //   led_aus(1);
    }
    if (func==3)db4|=(1<<2);
    if (func==2)db4|=(1<<1);
    if (func==1)db4|=(1<<0);
   if (fLok==218){
    a218db4=db4;
    
    sendLocoInfo(toParse[6],toParse[7], a218dirspeed, a218db4,0,0,0);
  }
  else
   sendLocoInfo(toParse[6],toParse[7], 0, db4,0,0,0);
  }
  if ((toParse[2]==0x040)&&(toParse[3]==0x00)&&(toParse[4]==0xe4)&&(toParse[5]==0x13)){
    int myLok=toParse[7];
    if(toParse[6]>192) myLok=((toParse[6]-192)*256)+toParse[7];
    locoSpdMsg mlok;
    mlok.command=0xa0; //Speed-Message eintragen!
    int adr_hi=myLok/128;
    int adr_lo=myLok%128;
    int temp2=toParse[8];
  
    int myslot=ret_slot(adr_lo, adr_hi); //falscher Slot, Fehler suchen!
  if (myLok==218){
    a218dirspeed=temp2;
    tempdir=0;
    if (a218dirspeed & (1<<7)) tempdir=1;
    if (a218dir != tempdir){
      a218dir=tempdir;
      if (a218dir==1){
        a218dirspeed = 0x80;
      }
      else a218dirspeed=0;
    }
    motor_run(a218dirspeed);
    sendLocoInfo(toParse[6],toParse[7], a218dirspeed, a218db4,0,0,0);
  }
  else sendLocoInfo(adr_hi,adr_lo, temp2, 0,0,0,0);
  }
}


boolean searchIP(IPAddress ip){
  for (int i=0; i < maxClients; i++){
    if (iplist[i].IP==ip) return true;
  }
  return false;
}




void loop() {
  byte packetSize = Udp.parsePacket();
  byte data[packetSize];
  if (packetSize) {
   IPAddress remoteIp = Udp.remoteIP();
   if (!searchIP(remoteIp)){
    if (ipcounter<maxClients){
      iplist[ipcounter].IP=remoteIp;
      iplist[ipcounter].inuse=true;
      ipcounter++;
    }
   }
   Serial.print(remoteIp); 
   Serial.print(" : ");
   char packetBuffer[packetSize]; 
   byte len = Udp.read(packetBuffer, packetSize);
   if (len > 0) {
      packetBuffer[len] = 0;
    }
   for (int i=0; i<packetSize; i++){
    data[i]=byte(packetBuffer[i]); 
   }
     
  }
  
 if (packetSize>3)parseMessage(data,packetSize);
 if ((lesezeiger+1)<schreibzeiger) handleSer();
}
